/* drawmap v0.4 (April 2017)
 * Copyright (C) 2006, 2017 Hubai Tamas <htamas2@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see [ www.gnu.org/licenses/ ].
 */

/*========== Includes ==========*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#if defined WIN32 || _WIN32 || WIN64 || _WIN64
#include <windows.h>
#endif

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wstrict-prototypes"
#include <wand/magick-wand.h>
#pragma GCC diagnostic pop
/*========== Includes ==========*/

/*========== Defines ==========*/
#if defined WIN32 || _WIN32 || WIN64 || _WIN64
#define SLASH "\\"
#else
#define SLASH "/"
#endif

#define EXIT_NORMAL 0
#define EXIT_ERROR -1
#define PROG_NAME "drawmap"
#define PROG_VERSION "0.4"
#define COPYRIGHT "Copyright (C) 2006, 2017 Hubai Tamas"
#define MAX_PATHFILE 500
#define MAX_OPTION 100
#define GRID_COLOR_TILES "#050"
#define GRID_COLOR_ROOMS "#0f0"
#define TILES 30

#define ThrowWandException(wand) \
{ \
	char \
		*description; \
 \
	ExceptionType \
		severity; \
 \
	description = MagickGetException (wand, &severity); \
	(void) fprintf (stderr, "%s %s %ld %s\n", GetMagickModule(), description); \
	description = (char *)MagickRelinquishMemory (description); \
	exit (EXIT_ERROR); \
}

#define Check(ret) \
if (ret == MagickFalse) \
	ThrowWandException (mw)
/*========== Defines ==========*/

struct plvhead1 {
	char magic[7];
	char popversion;
	char plvversion;
	char lvlnum;
} head1;

struct plvhead2 {
	int numfields;
	int lvlsize;
} head2;

char *data;

/*** pointers & information about maps ***/
char *nbh[32];
char l1[32];
char ln[32];
char c1[32];
char cn[32];
char active[33];

/*** which pop version ***/
char popv;

int iFromWidth, iFromHeight, iToWidth, iToHeight;

MagickWand *mw; PixelWand *pw; DrawingWand *dw;
MagickWand *ow[10 + 2]; MagickWand *tw; MagickWand *rw;

/*** Command-line options. ***/
int iDimensions;
char cSet;
int iNoGrid;
int cEnvironment;

int Tile (int room, int tile);
int Background (int room, int tile, int byte);
int PopLink (int room, int dir);
int StartRoom (void);
void WriteVal (int m, int l, int c, unsigned char r);
void CheckBack (int a, int d, int r);
void Expand (int m, int l, int c);
void NewMap (int m, char r);
void FreeMap (int m);
void PNum (char n);
void PrintMap (int m);
void GenMaps (void);
void DrawRoom (int r);
void CheckAccess (char *sPathFile);
void GetOptionValue (char *sArgv, char *sValue);
void ShowUsage (void);

/*****************************************************************************/
int Tile (int room, int tile)
/*****************************************************************************/
{
	/*** room number: 1..32, tile number: 0..29 ***/
	return ((room - 1) * TILES + tile);
}
/*****************************************************************************/
int Background (int room, int tile, int byte)
/*****************************************************************************/
{
	/*** This function is currently not in use. ***/

	if (popv == 1)
	{
		return (720 + (room - 1) * TILES + tile + byte);
	} else {
		return (960 + (room - 1) * TILES * 4 + tile * 4 + byte);
	}
}
/*****************************************************************************/
int PopLink (int room, int dir)
/*****************************************************************************/
{
	/*** left 0, right 1, up 2, down 3 ***/

	if (popv == 1)
	{
		return (1952 + 4 * (room - 1) + dir);
	} else {
		return (6080 + 4 * (room - 1 ) + dir);
	}
}
/*****************************************************************************/
int StartRoom (void)
/*****************************************************************************/
{
	if (popv == 1)
	{
		return (2112);
	} else {
		return (6240);
	}
}
/*****************************************************************************/
void WriteVal (int m, int l, int c, unsigned char r)
/*****************************************************************************/
{
	char pr;

	if (r == 0) { return; }
	if ((l < 0) || (l > 70) || (c < 0) || (c > 70))
	{
		fprintf (stderr, "[FAILED] Edge of the map!\n");
		exit (EXIT_ERROR);
	}
	pr = nbh[m][l * 71 + c];
	if (pr == 0)
	{
		if (active[r] != 1)
		{
			fprintf (stderr, "[FAILED] Strange map geometry!\n");
			exit (EXIT_ERROR);
		}
		nbh[m][l * 71 + c] = r;
		if (l < l1[m]) { l1[m] = l; }
		if (l > ln[m]) { ln[m] = l; }
		if (c < c1[m]) { c1[m] = c; }
		if (c > cn[m]) { cn[m] = c; }
	} else if (pr != r) {
		fprintf (stderr, "[FAILED] Map violates symmetry rules [1]!\n");
		exit (EXIT_ERROR);
	}
}
/*****************************************************************************/
void CheckBack (int a, int d, int r)
/*****************************************************************************/
{
	if (a == 0) { return; }
	if (data[PopLink ((int)a, d)] != r)
	{
		fprintf (stderr, "[FAILED] Map violates symmetry rules [2]!\n");
		exit (EXIT_ERROR);
	}
}
/*****************************************************************************/
void Expand (int m, int l, int c)
/*****************************************************************************/
{
	unsigned char r;
	char a;

	if ((l < 0) || (l > 70) || (c < 0) || (c > 70))
	{
		fprintf (stderr, "[FAILED] Edge of the map!\n");
		exit (EXIT_ERROR);
	}
	r = nbh[m][l * 71 + c];
	if (r == 0) { return; }
	if (active[r] != 1) { return; }
	active[r] = 0;

	/*** left ***/
	a = data[PopLink ((int)r, 0)];
	CheckBack (a, 1, r);
	WriteVal (m, l, c - 1, a);
	Expand (m, l, c - 1);

	/*** right ***/
	a = data[PopLink ((int)r, 1)];
	CheckBack (a, 0, r);
	WriteVal (m, l, c + 1, a);
	Expand (m, l, c + 1);

	/*** up ***/
	a = data[PopLink ((int)r, 2)];
	CheckBack (a, 3, r);
	WriteVal (m, l - 1, c, a);
	Expand (m, l - 1, c);

	/*** down ***/
	a = data[PopLink ((int)r, 3)];
	CheckBack (a, 2, r);
	WriteVal (m, l + 1, c, a);
	Expand (m, l + 1, c);
}
/*****************************************************************************/
void NewMap (int m, char r)
/*****************************************************************************/
{
	nbh[m] = calloc (5041, sizeof (char));
	if (!nbh[m])
	{
		fprintf (stderr, "[FAILED] Out of memory!\n");
		exit (EXIT_ERROR);
	}
	nbh[m][2520] = r;
	l1[m] = 35;
	ln[m] = 35;
	c1[m] = 35;
	cn[m] = 35;
	Expand (m, 35, 35);
}
/*****************************************************************************/
void FreeMap (int m)
/*****************************************************************************/
{
	free (nbh[m]);
}
/*****************************************************************************/
void PNum (char n)
/*****************************************************************************/
{
	if (n == 0)
	{
		printf (".. ");
	} else if (n < 10) {
		printf (".%hhd ", n);
	} else {
		printf ("%hhd ", n);
	}
}
/*****************************************************************************/
void PrintMap (int m)
/*****************************************************************************/
{
	/*** This function is currently not in use. ***/

	int l, c;

	if ((l1[m] == 35) && (ln[m] == 35) && (c1[m] == 35) && (cn[m] == 35))
	{
		/*** isolated room ***/
		active[(int)nbh[m][2520]] = 1;
		return;
	}
	printf ("[map component %d]\n", m);
	for (l = l1[m]; l <= ln[m]; l++)
	{
		for (c = c1[m]; c <= cn[m]; c++)
		{
			PNum (nbh[m][l * 71 + c]);
		}
		printf ("\n");
	}
	printf ("\n");
}
/*****************************************************************************/
void GenMaps (void)
/*****************************************************************************/
{
	int i, m;

	m = 0;

	for (i = 1; i <= 32; i++)
	{
		active[i] = 1;
	}
	i = data[StartRoom()];
/***
	for (i = 1; i <= 32; i++)
	{
***/
		if (active[i])
		{
			NewMap (m, i);
			m++;
		}
/***
	}
	for (i = 0; i < m; i++)
	{
		PrintMap (i);
	}
	printf ("isolated rooms: ");
	for (i = 1; i <= 32; i++)
	{
		if (active[i])
		{
			if (c) { printf (","); }
			printf ("%hhd", i);
			c++;
		}
	}
	if (!c) { printf ("none"); }
	printf ("\n");
	for (i = 0; i < m; i++)
	{
		FreeMap (i);
	}
***/
	for (i = 1; i < m; i++)
	{
		FreeMap (i);
	}
}
/*****************************************************************************/
void DrawRoom (int r)
/*****************************************************************************/
{
	/* draws the room in r to rw (room wand)
	 * uses pw, dw, tw (temp wand), ow[] (overlay wand)
	 * these (including rw) must be preinitialized
	 */
	int i, j, si, mod, tdi, di;
	char buf[8 + 2];

	if (r == 0)
	{
		Check (PixelSetColor (pw, "black")); /*** No room. ***/
	} else {
		Check (PixelSetColor (pw, "black"));
	}
	Check (MagickNewImage (rw, (iToWidth + 1) * 10, (iToHeight + 1) * 3, pw));

	ClearDrawingWand (dw);
	DrawSetStrokeWidth (dw, 1);

	if (r != 0)
	{
		printf ("Drawing room %d...\n", r);

		/*** place the objects ***/
		for (tdi = 0; tdi < TILES; tdi++)
		{
			if (tdi <= 9) { di = tdi + 20; }
				else if (tdi >= 20) { di = tdi - 20; }
					else { di = tdi; }
			si = data[Tile (r, di)];
			if (popv == 1)
			{
				/*** For stuck loose, use 0x1f. ***/
				if (si == 0x2b) { si = 0x1f; } else { si &= 0x1f; }
				if (iDimensions == 3)
				{
					mod = data[720 + Tile (r, di)];
					if (mod == -1) { mod = 10; } /*** 0xff ***/
					if ((si == 6) || (si == 15)) { mod = 0; } /*** buttons ***/
					if ((si == 4) && (mod == 2)) { mod = 0; } /*** closed gate ***/
				} else {
					mod = 0x00;
				}
			} else {
				if (si > 0x2f) { si = 0x2f; }
				mod = 0x00;
			}
			tw = MagickGetImageRegion (
				ow[mod],
				iFromWidth,
				iFromHeight,
				1 + (si % 8) * (iFromWidth + 1),
				1 + (si / 8) * (iFromHeight + 1)
			);
			MagickCompositeImage (
				rw,
				tw,
				OverCompositeOp,
				1 + (di % 10) * (iToWidth + 1),
				1 + (di / 10) * (iToHeight + 1) - (iFromHeight - iToHeight)
			);
		}

		/*** Draw the grid inside the rooms. ***/
		if (iNoGrid == 0)
		{
			Check (PixelSetColor (pw, GRID_COLOR_TILES));
			DrawSetStrokeColor (dw, pw);
			/*** Horizontal lines between tiles. ***/
			for (i = 0; i < 3; i++)
			{
				DrawLine (
					dw,
					0,
					i * (iToHeight + 1),
					((iToWidth + 1) * 10) - 1,
					i * (iToHeight + 1)
				);
			}
			/*** Vertical lines between tiles. ***/
			for (j = 0; j < 10; j++)
			{
				DrawLine (
					dw,
					j * (iToWidth + 1),
					0,
					j * (iToWidth + 1),
					(iToHeight * 3) + 2
				);
			}
		}
	}

	/* Draw the grid ABOVE and to the LEFT of the room.
	 * Outside the if() statement above, because blank screens should
	 * also get the grid.
	 */
	if (iNoGrid == 0)
	{
		Check (PixelSetColor (pw, GRID_COLOR_ROOMS));
		DrawSetStrokeColor (dw, pw);
		DrawLine (dw, 0, 0, ((iToWidth + 1) * 10) - 1, 0); /*** horizontal ***/
		DrawLine (dw, 0, 0, 0, ((iToHeight + 1) * 3) - 1); /*** vertical ***/
	}

	/*** Show room number. ***/
	if ((r != 0) && (iNoGrid == 0))
	{
		snprintf (buf, 8, "%d", r);
/***
		DrawSetStrokeAlpha (dw, 0.5);
		DrawSetFillAlpha (dw, 0.5);
***/
		DrawAnnotation (dw, 8, 16, (unsigned char *)buf);
/***
		DrawSetStrokeAlpha (dw, 1);
		DrawSetFillAlpha (dw, 1);
***/
	}

	Check (MagickDrawImage (rw, dw));
}
/*****************************************************************************/
void CheckAccess (char *sPathFile)
/*****************************************************************************/
{
	if (access (sPathFile, R_OK) == -1)
	{
		printf ("[FAILED] Could not open \"%s\": %s\n",
			sPathFile, strerror (errno));
		exit (EXIT_ERROR);
	}
}
/*****************************************************************************/
void GetOptionValue (char *sArgv, char *sValue)
/*****************************************************************************/
{
	int iTemp;
	char sTemp[MAX_OPTION + 2];

	iTemp = strlen (sArgv) - 1;
	snprintf (sValue, MAX_OPTION, "%s", "");
	while (sArgv[iTemp] != '=')
	{
		snprintf (sTemp, MAX_OPTION, "%c%s", sArgv[iTemp], sValue);
		snprintf (sValue, MAX_OPTION, "%s", sTemp);
		iTemp--;
	}
}
/*****************************************************************************/
void ShowUsage (void)
/*****************************************************************************/
{
	printf ("%s %s\n%s\n\n", PROG_NAME, PROG_VERSION, COPYRIGHT);
	printf ("Usage:\n");
	printf ("  %s [OPTIONS] <PLV file>\n\nOptions:\n", PROG_NAME);
	printf ("  -h, -?,    --help              display this help and exit\n");
	printf ("  -v,        --version           output version information and"
		" exit\n");
	printf ("  -d=DIM,    --dimensions=DIM    use DIM dimensions\n");
	printf ("  -s=SET,    --set=SET           use tile graphics SET\n");
	printf ("  -n,        --nogrid            do not add the grid or"
		" room numbers\n");
	printf ("  -e=ENV,    --environment=ENV   use ENV\n\n");
	printf ("Dimensions: 2, 3 [default = 2]\n");
	printf ("Sets: a [default = a]\n");
	printf ("Environments: g, d, p [default = g]\n\n");
	printf ("Important: DIM 3 must be used with either ENV d or p.\n");
	printf ("Currently, DIM 3 only works with PoP1.\n");
	printf ("\n");
	exit (EXIT_NORMAL);
}
/*****************************************************************************/
int main (int argc, char *argv[])
/*****************************************************************************/
{
	FILE *f;
	int r, l, c;
	char cks;
	int width, height;
	int x, y;
	int xs; int xt; int ys; int yt;
	char sDATFile[MAX_PATHFILE + 2];
	char sOptionValue[MAX_OPTION + 2];
	char sPNG[MAX_PATHFILE + 2];
	int iMaxMod, iImageMod;

	/*** Used for looping. ***/
	int iArgLoop;
	int iModLoop;

	/*** Defaults. ***/
	iDimensions = 2; /*** 2D ***/
	cSet = 'a';
	iNoGrid = 0;
	cEnvironment = 'g'; /*** general ***/

	if (argc > 1)
	{
		for (iArgLoop = 1; iArgLoop <= argc - 1; iArgLoop++)
		{
			if ((strcmp (argv[iArgLoop], "-h") == 0) ||
				(strcmp (argv[iArgLoop], "-?") == 0) ||
				(strcmp (argv[iArgLoop], "--help") == 0))
			{
				ShowUsage();
			}
			else if ((strcmp (argv[iArgLoop], "-v") == 0) ||
				(strcmp (argv[iArgLoop], "--version") == 0))
			{
				printf ("%s %s\n", PROG_NAME, PROG_VERSION);
				exit (EXIT_NORMAL);
			}
			else if ((strncmp (argv[iArgLoop], "-d=", 3) == 0) ||
				(strncmp (argv[iArgLoop], "--dimensions=", 13) == 0))
			{
				GetOptionValue (argv[iArgLoop], sOptionValue);
				iDimensions = atoi (sOptionValue);
			}
			else if ((strncmp (argv[iArgLoop], "-s=", 3) == 0) ||
				(strncmp (argv[iArgLoop], "--set=", 6) == 0))
			{
				GetOptionValue (argv[iArgLoop], sOptionValue);
				cSet = sOptionValue[0];
			}
			else if ((strcmp (argv[iArgLoop], "-n") == 0) ||
				(strcmp (argv[iArgLoop], "--nogrid") == 0))
			{
				iNoGrid = 1;
			}
			else if ((strncmp (argv[iArgLoop], "-e=", 3) == 0) ||
				(strncmp (argv[iArgLoop], "--environment=", 14) == 0))
			{
				GetOptionValue (argv[iArgLoop], sOptionValue);
				cEnvironment = sOptionValue[0];
			}
			else
			{
				snprintf (sDATFile, MAX_PATHFILE, "%s", argv[iArgLoop]);
			}
		}
	} else {
		ShowUsage();
	}

	f = fopen (sDATFile, "rb");
	if (f == NULL)
	{
		fprintf (stdout, "[FAILED] Unable to open \"%s\"!\n", sDATFile);
		exit (EXIT_ERROR);
	}

	/*** Magic ID (7), POP Version (1), PLV Version (1), Level Number (1). ***/
	fread (&head1, 10, 1, f);

	/*** Number Of Fields (4), Level Size (4). ***/
	fread (&head2, 8, 1, f);

	/*** Checksum. ***/
	fread (&cks, 1, 1, f);

	data = malloc (head2.lvlsize);
	if (!data)
	{
		fprintf (stdout, "[FAILED] Out of memory!\n");
		exit (EXIT_ERROR);
	}
	fread (data, head2.lvlsize, 1, f);

	popv = head1.popversion;
	switch (iDimensions)
	{
		case 2: /*** 2D ***/
			iFromWidth = 16;
			iFromHeight = 32;
			iToWidth = 16 - iNoGrid;
			iToHeight = 32 - iNoGrid;
			iMaxMod = 0;
			break;
		case 3: /*** 3D ***/
			iFromWidth = 25;
			iFromHeight = 32;
			iToWidth = 14 - iNoGrid;
			iToHeight = 26 - iNoGrid;
			iMaxMod = 10;
			break;
		default:
			printf ("[FAILED] Unknown dimensions: %i!\n", iDimensions);
			exit (EXIT_ERROR);
	}

	GenMaps();
	xs = c1[0];
	xt = cn[0];
	ys = l1[0];
	yt = ln[0];

	MagickWandGenesis();
	mw = NewMagickWand();
	pw = NewPixelWand();
	dw = NewDrawingWand();
	for (iModLoop = 0; iModLoop <= iMaxMod; iModLoop++)
	{
		iImageMod = iModLoop;
		if (iImageMod == 10) { iImageMod = 0xff; }
		ow[iModLoop] = NewMagickWand();
		snprintf (sPNG, MAX_PATHFILE, "png%sPoP%i_%iD%c%s%c%02x.png",
			SLASH, popv, iDimensions, cSet, SLASH, cEnvironment, iImageMod);
		CheckAccess (sPNG);
		Check (MagickReadImage (ow[iModLoop], sPNG));
	}
	tw = NewMagickWand();
	rw = NewMagickWand();

	Check (PixelSetColor (pw, "white"));
	width = (xt - xs + 1) * ((iToWidth + 1) * 10) + 1 - iNoGrid;
	height = (yt - ys + 1) * ((iToHeight + 1) * 3) + 1 - iNoGrid;
	Check (MagickNewImage (mw, width, height, pw));

	for (l = ys; l <= yt; l++)
	{
		y = (l - ys) * ((iToHeight + 1) * 3);
		for (c = xs; c <= xt; c++)
		{
			r = nbh[0][l * 71 + c];
			x = (c - xs) * ((iToWidth + 1) * 10);
			DrawRoom (r);
			MagickCompositeImage (mw, rw, OverCompositeOp, x, y);
		}
	}

	/*** Draw the grid lines far right and bottom. ***/
	if (iNoGrid == 0)
	{
		ClearDrawingWand (dw);
		Check (PixelSetColor (pw, GRID_COLOR_ROOMS));
		DrawSetStrokeColor (dw, pw);
/***
		DrawSetStrokeAlpha (dw, 1);
		DrawSetFillAlpha (dw, 1);
***/
		DrawLine (dw, width - 1, 0, width - 1, height - 1);
		DrawLine (dw, 0, height - 1, width - 1, height - 1);
	}

	Check (MagickDrawImage (mw, dw));

	Check (MagickWriteImage (mw, "drawmap.png"));
	mw = DestroyMagickWand (mw);
	pw = DestroyPixelWand (pw);
	dw = DestroyDrawingWand (dw);
	for (iModLoop = 0; iModLoop <= iMaxMod; iModLoop++)
		{ ow[iModLoop] = DestroyMagickWand (ow[iModLoop]); }
	tw = DestroyMagickWand (tw);
	MagickWandTerminus();

	fclose (f);
	return 0;
}
/*****************************************************************************/
